// QuickDraw 3d Sample Code
//
// This file contains utility routines for QuickDraw 3d sample code.
// Shows how to read a metafile and render it.
//
// Created 27th Dec 1994, Nick Thompson, DEVSUPPORT


#include <Files.h>
#include <QuickDraw.h>
#include <QDOffScreen.h>
#include <StandardFile.h>

#include <QD3D.h>
#include <QD3DDrawContext.h>
#include <QD3DRenderer.h>
#include <QD3DShader.h>
#include <QD3DCamera.h>
#include <QD3DLight.h>
#include <QD3DGeometry.h>
#include <QD3DTransform.h>
#include <QD3DGroup.h>
#include <QD3DMath.h>
#include <QD3DStorage.h>
#include <QD3DIO.h>

#include "DSpSV_QD3D.h"
#include "DSpSV_DSpSupport.h"
#include "DSpSV_PictTools.h"

/*
************************************************************************
** constants
************************************************************************
*/
#define kTexturePICT 128

#define kCameraTo (-1.0)
#define kBackPlaneZ ( (kCameraTo) - 3.0 )

#define kPlaneWidth		3.0
#define kPlaneHeight	3.0

/*
************************************************************************
** globals
************************************************************************
*/
TQ3TransformObject gBackPlaneTransform;
TQ3StorageObject gMemoryStorage;

/*
************************************************************************
** prototypes
************************************************************************
*/
void MyPixmapFromGWorld( GWorldPtr inGWorld, TQ3Pixmap *ioPixmap );
void MyStoragePixmapFromGWorld( GWorldPtr inGWorld, TQ3StoragePixmap *ioPixmap );
OSStatus MyCreatePixmapFromPICTResource( SInt16 inResourceID, TQ3StoragePixmap *outPixmap );

static	TQ3FileObject 		MyGetNewFile( FSSpec *myFSSpec, TQ3Boolean *isText ) ;

void GetGroupBBox(
	DocumentPtr			theDocument,
	TQ3BoundingBox 		*viewBBox) ;
												
static	TQ3Status MyAddShaderToGroup( TQ3GroupObject group ) ;

static TQ3Status GetDocumentGroupBoundingBox( 
	DocumentPtr theDocument , 
	TQ3BoundingBox *viewBBox) ;

//-----------------------------------------------------------------------------------------------
// Submit the scene for rendering/fileIO and picking
TQ3Status SubmitScene( DocumentPtr theDocument, Boolean inLeftView, Boolean inUseBackPlane ) 
{		
	TQ3Vector3D				globalScale;
	TQ3Vector3D				globalTranslate;
	TQ3ViewObject			theView;
	
	globalScale.x = globalScale.y = globalScale.z = theDocument->fGroupScale;
	globalTranslate = *(TQ3Vector3D *)&theDocument->fGroupCenter;
	Q3Vector3D_Scale(&globalTranslate, -1, &globalTranslate);

	/* choose eye view */	
	if( inLeftView )
		theView = theDocument->leftView;
	else
		theView = theDocument->rightView;

	/* --- global settings --- */
	
	/* submit global rendering styles */
	Q3Style_Submit(theDocument->fInterpolation, theView);
	Q3Style_Submit(theDocument->fBackFacing , theView);
	Q3Style_Submit(theDocument->fFillStyle, theView);
	
	/* --- back plane --- */

	/* submit back plane */
	if( theDocument->fBackPlane && inUseBackPlane )
	{
		Q3Push_Submit( theView );

		Q3Transform_Submit( gBackPlaneTransform, theView );
		Q3DisplayGroup_Submit( theDocument->fBackPlane, theView);
		
		Q3Pop_Submit( theView );
	}
	
	/* --- 3DMF model --- */

	/* submit model rotation */
	Q3MatrixTransform_Submit( &theDocument->fRotation, theView);
	
	/* submit model scale */
	Q3ScaleTransform_Submit(&globalScale, theView);
	
	/* submit model transform */
	Q3TranslateTransform_Submit(&globalTranslate, theView);
	
	/* submit model */
	Q3DisplayGroup_Submit( theDocument->fModel, theView);
	
	return kQ3Success ;
}

//-----------------------------------------------------------------------------------------------

static TQ3Status GetDocumentGroupBoundingBox( 
	DocumentPtr theDocument , 
	TQ3BoundingBox *viewBBox)
{
	TQ3Status		status;
	TQ3ViewStatus	viewStatus ;
	TQ3BoundingBox	leftBox, rightBox;
	
	status = Q3View_StartBoundingBox( theDocument->leftView, kQ3ComputeBoundsApproximate );
	do
	{
		status = SubmitScene( theDocument, true, false ) ;
	} while((viewStatus = Q3View_EndBoundingBox( theDocument->leftView, &leftBox )) == kQ3ViewStatusRetraverse );

	status = Q3View_StartBoundingBox( theDocument->rightView, kQ3ComputeBoundsApproximate );
	do
	{
		status = SubmitScene( theDocument, false, false ) ;
	} while((viewStatus = Q3View_EndBoundingBox( theDocument->rightView, &rightBox )) == kQ3ViewStatusRetraverse );

	*viewBBox = leftBox;
	if( viewBBox->max.x > rightBox.max.x )
		viewBBox->max.x = rightBox.max.x;
		
	if( viewBBox->max.y > rightBox.max.y )
		viewBBox->max.y = rightBox.max.y;
		
	if( viewBBox->max.z > rightBox.max.z )
		viewBBox->max.z = rightBox.max.z;
		
	if( viewBBox->min.x > rightBox.min.x )
		viewBBox->min.x = rightBox.min.x;
		
	if( viewBBox->min.y > rightBox.min.y )
		viewBBox->min.y = rightBox.min.y;
		
	if( viewBBox->min.z > rightBox.min.z )
		viewBBox->min.z = rightBox.min.z;
		
	return status ;
}

//-----------------------------------------------------------------------------------------------

TQ3ViewObject MyNewView(WindowPtr theWindow )
{
	TQ3Status				myStatus;
	TQ3ViewObject			myView;
	TQ3DrawContextObject		myDrawContext;
	TQ3RendererObject		myRenderer;
	TQ3CameraObject			myCamera;
	TQ3GroupObject			myLights;
	
	myView = Q3View_New();
	
	//	Create and set draw context.
	if ((myDrawContext = MyNewDrawContext(theWindow)) == nil )
		goto bail;
		
	if ((myStatus = Q3View_SetDrawContext(myView, myDrawContext)) == kQ3Failure )
		goto bail;

	Q3Object_Dispose( myDrawContext ) ;
	
	// Create and set renderer.
	//
	// hacky way to do this, but since I wanted these snippets to have 
	// a minimal interface, this will suffice
	//
	// change the next line to #if 1 to use the WF renderer
	
#if 0
	// this would use the wireframe renderer
	myRenderer = Q3Renderer_NewFromType(kQ3RendererTypeWireFrame);
	if ((myStatus = Q3View_SetRenderer(myView, myRenderer)) == kQ3Failure ) {
		goto bail;
	}
#else
	// this would use the interactive software renderer

	if ((myRenderer = Q3Renderer_NewFromType(kQ3RendererTypeInteractive)) != nil ) {
		Q3InteractiveRenderer_SetDoubleBufferBypass( myRenderer, kQ3True );
		if ((myStatus = Q3View_SetRenderer(myView, myRenderer)) == kQ3Failure ) {
			goto bail;
		}
	}
	else {
		goto bail;
	}
#endif

	Q3Object_Dispose( myRenderer ) ;
	
	//	Create and set camera.
	if ( (myCamera = MyNewCamera(theWindow)) == nil )
		goto bail;
		
	if ((myStatus = Q3View_SetCamera(myView, myCamera)) == kQ3Failure )
		goto bail;

	Q3Object_Dispose( myCamera ) ;
	
	//	Create and set lights.
	if ((myLights = MyNewLights()) == nil )
		goto bail;
		
	if ((myStatus = Q3View_SetLightGroup(myView, myLights)) == kQ3Failure )
		goto bail;
		
	Q3Object_Dispose(myLights);

	//	Done!!!
	return ( myView );
	
bail:
	//	If any of the above failed, then don't return a view.
	return ( nil );
}

//----------------------------------------------------------------------------------
#define PIXMAP 1
TQ3DrawContextObject MyNewDrawContext( WindowPtr theWindow )
{
	TQ3DrawContextData		myDrawContextData;
#if PIXMAP
	TQ3PixmapDrawContextData	myPixmapDrawContextData;
#else
	TQ3MacDrawContextData	myMacDrawContextData;
#endif
	TQ3ColorARGB			ClearColor;
	TQ3DrawContextObject	myDrawContext ;

	ClearColor.a = 1.0;
	ClearColor.r = 0;
	ClearColor.g = 0;
	ClearColor.b = 0;
	
	//	Fill in draw context data.
	myDrawContextData.clearImageMethod = kQ3ClearMethodWithColor;
	myDrawContextData.clearImageColor = ClearColor;
	
	myDrawContextData.paneState = kQ3False;
	myDrawContextData.maskState = kQ3False;
	
	myDrawContextData.doubleBufferState = kQ3False;

#if PIXMAP
	myPixmapDrawContextData.drawContextData = myDrawContextData;
	MyPixmapFromGWorld( (CGrafPtr)theWindow, &myPixmapDrawContextData.pixmap );

	//	Create draw context and return it, if its nil the caller must handle
	myDrawContext = Q3PixmapDrawContext_New( &myPixmapDrawContextData ) ;
#else
	myMacDrawContextData.drawContextData = myDrawContextData;
	myMacDrawContextData.window = (CGrafPtr) theWindow;		// this is the window associated with the view
	myMacDrawContextData.library = kQ3Mac2DLibraryNone;
	myMacDrawContextData.viewPort = nil;
	myMacDrawContextData.grafPort = nil;
	
	//	Create draw context and return it, if its nil the caller must handle
	myDrawContext = Q3MacDrawContext_New(&myMacDrawContextData) ;

#endif

	return myDrawContext ;
}

void MyPixmapFromGWorld(
	GWorldPtr inGWorld,
	TQ3Pixmap *ioPixmap
)
{
	PixMapHandle thePM;
	PixMap *theQDPixmap;

	thePM = GetGWorldPixMap( inGWorld );
	LockPixels( thePM );
	theQDPixmap = *thePM;
	
	/* assumes big endian, rgb 16 format */
	ioPixmap->image			= theQDPixmap->baseAddr;
	ioPixmap->width			= theQDPixmap->bounds.right - theQDPixmap->bounds.left;
	ioPixmap->height		= theQDPixmap->bounds.bottom - theQDPixmap->bounds.top;
	ioPixmap->rowBytes		= theQDPixmap->rowBytes & 0x3FFF;
	ioPixmap->pixelSize		= theQDPixmap->pixelSize;
	ioPixmap->pixelType		= kQ3PixelTypeRGB16;
	ioPixmap->bitOrder		= kQ3EndianBig;
	ioPixmap->byteOrder		= kQ3EndianBig;

	UnlockPixels( GetGWorldPixMap( inGWorld ) );
}

void MyStoragePixmapFromGWorld(
	GWorldPtr inGWorld,
	TQ3StoragePixmap *ioPixmap
)
{
	UInt32 theSize;
	PixMapHandle thePM;
	PixMap *theQDPixmap;

	thePM = GetGWorldPixMap( inGWorld );
	LockPixels( thePM );
	theQDPixmap = *thePM;
	
	/* assumes big endian, rgb 16 format */
	ioPixmap->width			= theQDPixmap->bounds.right - theQDPixmap->bounds.left;
	ioPixmap->height		= theQDPixmap->bounds.bottom - theQDPixmap->bounds.top;
	ioPixmap->rowBytes		= theQDPixmap->rowBytes & 0x3FFF;
	ioPixmap->pixelSize		= theQDPixmap->pixelSize;
	ioPixmap->pixelType		= kQ3PixelTypeRGB16;
	ioPixmap->bitOrder		= kQ3EndianBig;
	ioPixmap->byteOrder		= kQ3EndianBig;
	
	/* have QD3D copy the image buffer into its own memory */
	/* use Q3MemoryStorage_NewBuffer() if you want to have it reference the memory instead */
	theSize					= ioPixmap->rowBytes * ioPixmap->height;
	ioPixmap->image			= Q3MemoryStorage_New( (UInt8 *)theQDPixmap->baseAddr, theSize );
	gMemoryStorage = ioPixmap->image; // hack, for deletion of object
	
	UnlockPixels( GetGWorldPixMap( inGWorld ) );
}

//----------------------------------------------------------------------------------

TQ3CameraObject MyNewCamera(WindowPtr theWindow)
{
	TQ3CameraObject					myCamera;
	TQ3CameraData					myCameraData;
	TQ3ViewAngleAspectCameraData	myViewAngleCameraData;
	
	TQ3ViewPlaneCameraData			myViewPlaneCameraData;
	
	TQ3Point3D						cameraFrom 	= { 0.0, 0.0, 30.0 };
	TQ3Point3D						cameraTo 	= { 0.0, 0.0, kCameraTo };
	TQ3Vector3D						cameraUp 	= { 0.0, 1.0, 0.0 };
	
	float 							fieldOfView = .52359333333;
	float 							hither 		= 0.001;
	float 							yon 		= 1000;
	
	//	Fill in camera data.
	myCameraData.placement.cameraLocation = cameraFrom;
	myCameraData.placement.pointOfInterest = cameraTo;
	myCameraData.placement.upVector = cameraUp;
	
	myCameraData.range.hither = hither;
	myCameraData.range.yon = yon;
	
	myCameraData.viewPort.origin.x = -1.0;
	myCameraData.viewPort.origin.y = 1.0;
	myCameraData.viewPort.width = 2.0;
	myCameraData.viewPort.height = 2.0;
	
	myViewPlaneCameraData.cameraData = myCameraData;
	myViewPlaneCameraData.viewPlane = 20.0;
	myViewPlaneCameraData.halfWidthAtViewPlane = 1.0;
	myViewPlaneCameraData.halfHeightAtViewPlane = 1.0;
	myViewPlaneCameraData.centerXOnViewPlane = 0.0;
	myViewPlaneCameraData.centerYOnViewPlane = 0.0;
	
	myViewAngleCameraData.cameraData = myCameraData;
	myViewAngleCameraData.fov = fieldOfView ;
	
	// set up the aspect ratio based on the window
	myViewAngleCameraData.aspectRatioXToY =  
			(float) (theWindow->portRect.right - theWindow->portRect.left) / 
			(float) (theWindow->portRect.bottom - theWindow->portRect.top);

/*	myCamera = Q3ViewAngleAspectCamera_New(&myViewAngleCameraData);	
*/	
	myCamera = Q3ViewPlaneCamera_New(&myViewPlaneCameraData);	
	//	Return the camera.
	return ( myCamera );
}


//----------------------------------------------------------------------------------

TQ3GroupObject MyNewLights()
{
	TQ3GroupPosition			myGroupPosition;
	TQ3GroupObject			myLightList;
	TQ3LightData				myLightData;
	TQ3PointLightData		myPointLightData;
	TQ3DirectionalLightData	myDirectionalLightData;
	TQ3LightObject			myAmbientLight, myPointLight, myFillLight;
	TQ3Point3D				pointLocation = { -10.0, 0.0, 10.0 };
	TQ3Vector3D				fillDirection = { 10.0, 0.0, 10.0 };
	TQ3ColorRGB				WhiteLight = { 1.0, 1.0, 1.0 };
	
	//	Set up light data for ambient light.  This light data will be used for point and fill
	//	light also.

	myLightData.isOn = kQ3True;
	myLightData.color = WhiteLight;
	
	//	Create ambient light.
	myLightData.brightness = .2;
	myAmbientLight = Q3AmbientLight_New(&myLightData);
	if ( myAmbientLight == nil )
		goto bail;
	
	//	Create point light.
	myLightData.brightness = 1.0;
	myPointLightData.lightData = myLightData;
	myPointLightData.castsShadows = kQ3False;
	myPointLightData.attenuation = kQ3AttenuationTypeNone;
	myPointLightData.location = pointLocation;
	myPointLight = Q3PointLight_New(&myPointLightData);
	if ( myPointLight == nil )
		goto bail;

	//	Create fill light.
	myLightData.brightness = .2;
	myDirectionalLightData.lightData = myLightData;
	myDirectionalLightData.castsShadows = kQ3False;
	myDirectionalLightData.direction = fillDirection;
	myFillLight = Q3DirectionalLight_New(&myDirectionalLightData);
	if ( myFillLight == nil )
		goto bail;

	//	Create light group and add each of the lights into the group.
	myLightList = Q3LightGroup_New();
	if ( myLightList == nil )
		goto bail;
	myGroupPosition = Q3Group_AddObject(myLightList, myAmbientLight);
	if ( myGroupPosition == 0 )
		goto bail;
	myGroupPosition = Q3Group_AddObject(myLightList, myPointLight);
	if ( myGroupPosition == 0 )
		goto bail;
	myGroupPosition = Q3Group_AddObject(myLightList, myFillLight);
	if ( myGroupPosition == 0 )
		goto bail;

	Q3Object_Dispose( myAmbientLight ) ;
	Q3Object_Dispose( myPointLight ) ;
	Q3Object_Dispose( myFillLight ) ;

	//	Done!
	return ( myLightList );
	
bail:
	//	If any of the above failed, then return nothing!
	return ( nil );
}

//----------------------------------------------------------------------------------

TQ3GroupObject MyNewModelFromFile(FSSpec *theFileSpec)
{
	TQ3GroupObject		myGroup = NULL ;
	TQ3Boolean			isText = kQ3False ;
	TQ3FileMode			myFileMode = 0;
	TQ3FileObject		theFile;
	
	//	Create a ordered group for the complete model.
	if ((myGroup = Q3DisplayGroup_New()) == NULL )
		return NULL;
		
	MyAddShaderToGroup( myGroup ) ;

	theFile = MyGetNewFile( theFileSpec, &isText ) ;
	
	if( isText == kQ3True )
		myFileMode |= kQ3FileModeText;	// is it a text metafile??	

	// Open the file object
	if( Q3File_OpenRead( theFile, &myFileMode ) != kQ3Success)
		return  NULL ;

	if( MyReadModelFromFile( theFile, myGroup ) == 0)
		DebugStr("\pMetafile data read is null") ;
	
	Q3File_Close(theFile);			// close and dispose of the file object
	Q3Object_Dispose(theFile);
	
	return myGroup ;
}

OSStatus
MyCreatePixmapFromPICTResource(
	SInt16 inResourceID,
	TQ3StoragePixmap *outPixmap
)
{
	OSStatus theError = noErr;
	GWorldPtr theGWorld;
		
	/* load the PICT into a GWorld */
	theError = LoadPictIntoGWorld( inResourceID, kDisplayDepth, 0, 0, &theGWorld );
	
	/* wrap the pixmap around the gworld */
	if( noErr == theError )
		MyStoragePixmapFromGWorld( theGWorld, outPixmap );
	if( outPixmap->image == NULL )
		theError = memFullErr;
		
	/* done */
	if( theGWorld )
		DisposeGWorld( theGWorld );

	return theError;
}

TQ3GroupObject MyNewBackPlane(
	DocumentPtr /* inDocument */
)
{
	Boolean theError = false;
	TQ3GroupObject theGroup = NULL ;
	TQ3ShaderObject theShader = NULL;
	TQ3TextureObject theTexture = NULL;
	TQ3StoragePixmap thePixmap;
	TQ3RotateTransformData theRotationData = { kQ3AxisZ, 0 };
	
	/* create the rotation transform */
	gBackPlaneTransform = Q3RotateTransform_New( &theRotationData );
	if( NULL == gBackPlaneTransform )
		theError = true;
	
	/* create the ordered display group */
	theGroup = Q3DisplayGroup_New();
	if( NULL == theGroup )
		theError = true;
	
	/* create the pixmap */
	if( false == theError )
	{
		if( MyCreatePixmapFromPICTResource( kTexturePICT, &thePixmap ) != noErr )
			theError = true;
	}
	
	/* create the texture map */
	if( false == theError )
	{
		theTexture = Q3PixmapTexture_New( &thePixmap );
		if( NULL == theTexture )
			theError = true;
	}

	/* create a texture shader that uses the texture map */
	if( false == theError )
	{
		theShader = Q3TextureShader_New( theTexture );
		if( NULL == theShader )
			theError = true;
	}

	/* create the plane, Y axis is up */
	if( false == theError )
	{
		TQ3Vertex3D thePlaneVertices[4];
		TQ3PolygonData thePlane;
		TQ3GeometryObject thePolygon;
		TQ3Param2D	theUV;

		/* upper left */
		thePlaneVertices[0].point.x			= -(kPlaneWidth/2);
		thePlaneVertices[0].point.y			= -(kPlaneHeight/2);
		thePlaneVertices[0].point.z			= kBackPlaneZ;
		thePlaneVertices[0].attributeSet	= Q3AttributeSet_New();
		if( NULL == thePlaneVertices[0].attributeSet )
			theError = true;
		if( false == theError )
		{
			theUV.u = 0;
			theUV.v = 0;
			Q3AttributeSet_Add( thePlaneVertices[0].attributeSet, kQ3AttributeTypeShadingUV, &theUV );
		}
		
		/* upper right */
		if( false == theError )
		{
			thePlaneVertices[1].point.x			= (kPlaneWidth/2);
			thePlaneVertices[1].point.y			= -(kPlaneHeight/2);
			thePlaneVertices[1].point.z			= kBackPlaneZ;
			thePlaneVertices[1].attributeSet	= Q3AttributeSet_New();
			if( NULL == thePlaneVertices[1].attributeSet )
				theError = true;
			if( false == theError )
			{
				theUV.u = 1;
				theUV.v = 0;
				Q3AttributeSet_Add( thePlaneVertices[1].attributeSet, kQ3AttributeTypeShadingUV, &theUV );
			}
		}
		
		/* lower right */
		if( false == theError )
		{
			thePlaneVertices[2].point.x			= (kPlaneWidth/2);
			thePlaneVertices[2].point.y			= (kPlaneHeight/2);
			thePlaneVertices[2].point.z			= kBackPlaneZ;
			thePlaneVertices[2].attributeSet	= Q3AttributeSet_New();
			if( NULL == thePlaneVertices[2].attributeSet )
				theError = true;
			if( false == theError )
			{
				theUV.u = 1;
				theUV.v = 1;
				Q3AttributeSet_Add( thePlaneVertices[2].attributeSet, kQ3AttributeTypeShadingUV, &theUV );
			}
		}

		/* lower left */
		if( false == theError )
		{
			thePlaneVertices[3].point.x			= -(kPlaneWidth/2);
			thePlaneVertices[3].point.y			= (kPlaneHeight/2);
			thePlaneVertices[3].point.z			= kBackPlaneZ;
			thePlaneVertices[3].attributeSet	= Q3AttributeSet_New();
			if( NULL == thePlaneVertices[3].attributeSet )
				theError = true;
			if( false == theError )
			{
				theUV.u = 0;
				theUV.v = 1;
				Q3AttributeSet_Add( thePlaneVertices[3].attributeSet, kQ3AttributeTypeShadingUV, &theUV );
			}
		}
		
		/* init the polygon */
		thePlane.numVertices = 4;
		thePlane.vertices = thePlaneVertices;
		thePlane.polygonAttributeSet = Q3AttributeSet_New();
		if( NULL == thePlane.polygonAttributeSet )
			theError = true;

		/* create the polygon */
		if( false == theError )
		{
			thePolygon = Q3Polygon_New( &thePlane );
			if( NULL == thePolygon )
				theError = true;
			
			/* add the texture shader as an attribute */
			Q3AttributeSet_Add( thePlane.polygonAttributeSet, kQ3AttributeTypeSurfaceShader, &theShader );
			
			/* add it to the group */
			Q3Group_AddObject( theGroup, thePolygon );
		}
		
		/* handle errors */
		if( theError && thePolygon )
		{
			Q3Object_Dispose( thePolygon );
		}

		/* dispose of object references */
		if( thePlaneVertices[0].attributeSet )
			Q3Object_Dispose( thePlaneVertices[0].attributeSet );
		if( thePlaneVertices[1].attributeSet )
			Q3Object_Dispose( thePlaneVertices[1].attributeSet );
		if( thePlaneVertices[2].attributeSet )
			Q3Object_Dispose( thePlaneVertices[2].attributeSet );
		if( thePlaneVertices[3].attributeSet )
			Q3Object_Dispose( thePlaneVertices[3].attributeSet );
		if( thePlane.polygonAttributeSet )
			Q3Object_Dispose( thePlane.polygonAttributeSet );
		if( thePolygon )
			Q3Object_Dispose( thePolygon );
	}
		
	/* dispose of object references */
	if( theShader )
		Q3Object_Dispose( theShader );
	if( theTexture )
		Q3Object_Dispose( theTexture );

	/* only dispose of the group if there was an error */
	if( theError && theGroup )
		Q3Object_Dispose( theGroup );
	
	return theGroup;
}

//----------------------------------------------------------------------------------
// attach a shader to the group

TQ3Status MyAddShaderToGroup( TQ3GroupObject group )
{
	TQ3ShaderObject	illuminationShader = Q3PhongIllumination_New();

	Q3Group_AddObject(group, illuminationShader);
	Q3Object_Dispose(illuminationShader);
	return(kQ3Success);
}

//----------------------------------------------------------------------------------
// read model from file object into the supplied group

TQ3Status MyReadModelFromFile( TQ3FileObject theFile,TQ3GroupObject myGroup)
{	
	if(theFile != NULL) {
	
		TQ3Object			myTempObj ;
		TQ3Boolean			isEOF ;
				
	
		// read objects from the file
		do {
		
			myTempObj = Q3File_ReadObject( theFile );
			
			if( myTempObj != NULL ) {
				// we only want the object in our main group if we can draw it
				if ( Q3Object_IsDrawable( myTempObj) ) 
					Q3Group_AddObject( myGroup, myTempObj ) ;
				
				// we either added the object to the main group, or we don't care
				// so we can safely dispose of the object
				Q3Object_Dispose( myTempObj ) ;
			}
			
			// check to see if we reached the end of file yet
			isEOF = Q3File_IsEndOfFile( theFile );
			
		} while (isEOF == kQ3False);	
	}
	
	if( myGroup != NULL )
		return kQ3Success ;
	else
		return kQ3Failure ;
}

//-----------------------------------------------------------------------------------------------
// cleaned up from IM QuickDraw 3D pp 15-5
static TQ3FileObject MyGetNewFile( FSSpec *myFSSpec, TQ3Boolean *isText )
{
	TQ3FileObject		myFileObj;
	TQ3StorageObject		myStorageObj;
	OSType				myFileType;
	
	FInfo				fndrInfo ;

	// we assume the FSSpec passed in was valid, get the file information
	// we need to know the file type, this routine may get called by an appleEvent
	// handler, so we can't assume a type, we need to get it from the fsspec.
	
	FSpGetFInfo( myFSSpec, &fndrInfo ) ;
	
	// pull out the file type
	
	myFileType = fndrInfo.fdType ;
	
	// Create new storage object and new file object 
	if(((myStorageObj = Q3FSSpecStorage_New( myFSSpec )) == NULL) 
		|| ((myFileObj = Q3File_New()) == NULL)) 
	{
		if (myStorageObj != NULL) 
			Q3Object_Dispose(myStorageObj);
		return(NULL);
	}

	// Set the storage for the file object
	Q3File_SetStorage(myFileObj, myStorageObj);
	Q3Object_Dispose(myStorageObj);

	if (myFileType == '3DMF')
		*isText = kQ3False ;
	else if (myFileType == 'TEXT')
		*isText = kQ3True ;

	return (myFileObj);
}


//-------------------------------------------------------------------------------------------
//
Boolean MetafileFileSpecify( FSSpec *theFile )
{
	StandardFileReply	theSFReply ;
	SFTypeList			myTypes = { '3DMF' } ;
	const short			numTypes = 1 ;
		
	// Get the file name to open
	StandardGetFile( nil, numTypes, myTypes, &theSFReply ) ;
	
	if( theSFReply.sfGood )
		*theFile = theSFReply.sfFile ;
	
	// did the user cancel?
	return theSFReply.sfGood ;
	
}
//----------------------------------------------------------------------------------


void GetGroupBBox(
	DocumentPtr			theDocument,
	TQ3BoundingBox 		*viewBBox)
{
	TQ3Point3D 					from 	= { 0.0, 0.0, 1.0 };
	TQ3Point3D 					to 		= { 0.0, 0.0, 0.0 };
	TQ3Vector3D 				up 		= { 0.0, 1.0, 0.0 };
	
	float 						fieldOfView = .52359333333;
	float 						hither 		=  0.5;
	float 						yon 		=  1.5;
	TQ3GroupObject				mainGroup = theDocument->fModel ;

	TQ3Status					status;
	
	status = GetDocumentGroupBoundingBox( theDocument , viewBBox ) ;
								        
	//
	//  If we have a point model, then the "viewBBox" would end up
	//  being a "singularity" at the location of the point.  As
	//  this bounding "box" is used in setting up the camera spec,
	//  we get bogus input into Escher.
	
	{
 		float		xSize, ySize, zSize;
		
		xSize = viewBBox->max.x - viewBBox->min.x;
		ySize = viewBBox->max.y - viewBBox->min.y;
		zSize = viewBBox->max.z - viewBBox->min.z;

		if (xSize <= kQ3RealZero &&
		    ySize <= kQ3RealZero &&
			zSize <= kQ3RealZero) {
			
			viewBBox->max.x += 0.0001;
			viewBBox->max.y += 0.0001;
			viewBBox->max.z += 0.0001;
			
			viewBBox->min.x -= 0.0001;
			viewBBox->min.y -= 0.0001;
			viewBBox->min.z -= 0.0001;
		}
	}
}




//------------------------------------------------------------------------

TQ3Point3D AdjustCenterCamera( DocumentPtr theDocument );

TQ3Point3D AdjustCenterCamera( DocumentPtr theDocument )
{
	float 						fieldOfView;
	float 						hither;
	float 						yon;
	TQ3CameraPlacement			placement;
	TQ3CameraRange				range;
	TQ3BoundingBox 				viewBBox;
	long 						fromAxis;	
	float 						maxDimension;
 	float						xSize, ySize, zSize;
	float						weights[2] = { 0.5, 0.5 };
	TQ3Point3D					points[2];
	TQ3Vector3D				 	viewVector;
	TQ3Vector3D					normViewVector;
	TQ3Vector3D					eyeToFrontClip;
	TQ3Vector3D					eyeToBackClip;
	float						viewDistance;
	TQ3Vector3D					diagonalVector;
	float						ratio;
	TQ3CameraObject				camera;
	
	TQ3GroupObject				mainGroup = theDocument->fModel ;
	
	TQ3Point3D					*documentGroupCenter = &theDocument->fGroupCenter ;
	float						*documentGroupScale  = &theDocument->fGroupScale ;

	camera = theDocument->centerCamera;
	
	GetGroupBBox( theDocument, &viewBBox);

	/*
	 *  If we have a point model, then the "viewBBox" would end up
	 *  being a "singularity" at the location of the point.  As
	 *  this bounding "box" is used in setting up the camera spec,
	 *  we get bogus input into Escher.
	 */
	xSize = viewBBox.max.x - viewBBox.min.x;
	ySize = viewBBox.max.y - viewBBox.min.y;
	zSize = viewBBox.max.z - viewBBox.min.z;

	if (xSize <= kQ3RealZero &&
	    ySize <= kQ3RealZero &&
		zSize <= kQ3RealZero)  {
		viewBBox.max.x += 0.0001;
		viewBBox.max.y += 0.0001;
		viewBBox.max.z += 0.0001;
		
		viewBBox.min.x -= 0.0001;
		viewBBox.min.y -= 0.0001;
		viewBBox.min.z -= 0.0001;
	}

	points[0] = viewBBox.min;
	points[1] = viewBBox.max;

	Q3Point3D_AffineComb(points, weights, 2, documentGroupCenter);

	/*
	 *  The "from" point is on a vector perpendicular to the plane
	 *  in which the bounding box has greatest dimension.  As "up" is
	 *  always in the positive y direction, look at x and z directions.
	 */
	xSize = viewBBox.max.x - viewBBox.min.x;
	zSize = viewBBox.max.z - viewBBox.min.z;
	
	if (xSize > zSize) {
		fromAxis = kQ3AxisZ;
	} else {
		fromAxis = kQ3AxisX;
	}

	/*
	 *  Compute the length of the diagonal of the bounding box.
	 *
	 *  The hither and yon planes are adjusted so that the
 	 *  diagonal of the bounding box is 7/8 the size of the
 	 *  minimum dimension of the view frustum. The diagonal is used instead
 	 *  of the maximum size (in x, y, or z) so that when you rotate
 	 *  the object, the corners don't get clipped out.
 	 */
	Q3Point3D_Subtract(
		&viewBBox.max,
		&viewBBox.min,
		&diagonalVector);

	maxDimension	=	Q3Vector3D_Length(&diagonalVector);
	maxDimension	*=	8.0 / 7.0;
	
	ratio = 1.0 / maxDimension;
	ratio *= 3;
	
	*documentGroupScale = ratio;
	
	Q3Camera_GetPlacement(camera, &placement);

	Q3Point3D_Subtract(
		&placement.cameraLocation,
		&placement.pointOfInterest,
		&viewVector);
		
	viewDistance = Q3Vector3D_Length(&viewVector);
	
	Q3Vector3D_Normalize(&viewVector, &normViewVector);
	
	Q3Vector3D_Scale(&normViewVector, 
					 viewDistance - ratio * maxDimension/2.0,
					 &eyeToFrontClip);
					
	Q3Vector3D_Scale(&normViewVector, 
					viewDistance + ratio * maxDimension/2.0,
					&eyeToBackClip);

	hither 	= Q3Vector3D_Length(&eyeToFrontClip);
	yon 	= Q3Vector3D_Length(&eyeToBackClip);
	
	fieldOfView = 2 * atan((ratio * maxDimension/2.0)/hither);

	range.hither 				= hither;
	range.yon 					= yon;

	Q3Camera_SetRange(camera, &range);

//	Q3ViewAngleAspectCamera_SetFOV(
//		camera, fieldOfView);

//	Q3ViewAngleAspectCamera_SetAspectRatio(
//		camera, (float) winWidth / (float) winHeight);

//	Q3Object_Dispose(camera);
	
	return( *documentGroupCenter );
}

/*===========================================================================*\
 *
 *	Routine:	PositionCameras()
 *
 *	Comments:	
 *
\*===========================================================================*/
TQ3Status	PositionCameras( DocumentPtr theDocument )
{
	TQ3CameraPlacement	theCenterPlacement,
						theRightPlacement,
						theLeftPlacement;
	TQ3Matrix4x4		theWorldToViewMatrix;
	TQ3Vector3D			theCameraXAxis,
						theDisplacementVector;
	TQ3CameraObject		theCenterCamera;
	TQ3CameraObject		theRightCamera;
	TQ3CameraObject		theLeftCamera;

#if 0						
	float				theCenterViewPlane,
						theEyeViewPlane;
	float				theRatioOfViewPlanes;
	TQ3CameraViewPort	theCenterViewPort,
						theRightViewPort,
						theLeftViewPort;
	float				theXOfViewPortOrigin,
						theWidth,
						theHeight;
#endif

	AdjustCenterCamera( theDocument );
	
	theLeftCamera = theDocument->leftCamera;
	theRightCamera = theDocument->rightCamera;
	theCenterCamera = theDocument->centerCamera;
	
	if ((Q3Camera_GetPlacement(theCenterCamera, &theCenterPlacement))
		== kQ3Failure)
		return kQ3Failure;
	
	if ((Q3Camera_GetWorldToView(theCenterCamera, &theWorldToViewMatrix))
		== kQ3Failure)
		return kQ3Failure;
	
	/*	You can grab the cameraSpace x axis by doing this	*/
	
	Q3Vector3D_Set(&theCameraXAxis,
		theWorldToViewMatrix.value[0][0],
		theWorldToViewMatrix.value[0][1],
		theWorldToViewMatrix.value[0][2]);
	
	/*	Scale that normal to determine how much to shift the cameras	*/
	
	Q3Vector3D_Scale(&theCameraXAxis,
		theDocument->interaxialHalfDistance,
		&theDisplacementVector);
		
	/*	Shift Right camera	*/
	
	Q3Point3D_Vector3D_Add(&(theCenterPlacement.cameraLocation),
		&theDisplacementVector,
		&(theRightPlacement.cameraLocation));
	
/*	Q3Point3D_Vector3D_Add(&(theCenterPlacement.pointOfInterest),
		&theDisplacementVector,
		&(theRightPlacement.pointOfInterest));
*/		
	theRightPlacement.pointOfInterest = theCenterPlacement.pointOfInterest;

	theRightPlacement.upVector = theCenterPlacement.upVector;
		
	/*	Not sure if I can do this	*/
	Q3Vector3D_Negate(&theDisplacementVector, &theDisplacementVector);
	
	/*	Shift left camera	*/

	Q3Point3D_Vector3D_Add(&(theCenterPlacement.cameraLocation),
		&theDisplacementVector,
		&(theLeftPlacement.cameraLocation));
	
/*	Q3Point3D_Vector3D_Add(&(theCenterPlacement.pointOfInterest),
		&theDisplacementVector,
		&(theLeftPlacement.pointOfInterest));
*/	

	theLeftPlacement.pointOfInterest = theCenterPlacement.pointOfInterest;

	theLeftPlacement.upVector = theCenterPlacement.upVector;
		
	/*	We have to move the view planes of the cameras closer to the eye cameras
			*/

/*	Q3ViewPlaneCamera_GetViewPlane(theCenterCamera, &theCenterViewPlane);
	
	if (theCenterViewPlane == 0.0) return kQ3Failure;
	
	theEyeViewPlane	=
		 1 / (sin(atan((theDocument->interaxialHalfDistance + 1.0) / theCenterViewPlane)));
	
	Q3ViewPlaneCamera_SetViewPlane(theLeftCamera, theEyeViewPlane);
	Q3ViewPlaneCamera_SetViewPlane(theRightCamera, theEyeViewPlane);
	
	theRatioOfViewPlanes = (theEyeViewPlane / theCenterViewPlane);
	
	Q3Camera_GetViewPort(theCenterCamera, &theCenterViewPort);
	
	theWidth = theCenterViewPort.width * theRatioOfViewPlanes;
	theHeight = theCenterViewPort.height * theRatioOfViewPlanes;
	
	theXOfViewPortOrigin = theCenterViewPort.origin.x;
	
	theXOfViewPortOrigin = ((theXOfViewPortOrigin - 1.0) * theRatioOfViewPlanes) + 1.0;
	
	/*	for left eye	*/
	
/*	theLeftViewPort.origin.x	= theXOfViewPortOrigin;
	theLeftViewPort.origin.y	= theCenterViewPort.origin.y;
	theLeftViewPort.width		= theWidth;
	theLeftViewPort.height		= theHeight;
	
	Q3Camera_SetViewPort(theLeftCamera, &theLeftViewPort);
	
	/*	for right eye	*/
	
/*	theRightViewPort.origin.x	= - theXOfViewPortOrigin;
	theRightViewPort.origin.y	= theCenterViewPort.origin.y;
	theRightViewPort.width		= theWidth;
	theRightViewPort.height		= theHeight;
	
	Q3Camera_SetViewPort(theRightCamera, &theRightViewPort);
*/
	Q3Camera_SetPlacement(theLeftCamera, &theLeftPlacement);
	Q3Camera_SetPlacement(theRightCamera, &theRightPlacement);

	
	return kQ3Success;
}
